/************************************************************* 
			Mini Java 
 *************************************************************
 Auteur :	M. Sighireanu
 Fichier :	javaparser.mly
 Version :	1.2
 Date :		00/01/21 10:08:06
 *************************************************************
 Description :	parseur Caml de la syntaxe concrete Mini Java
 *************************************************************/

%{
open Javasyntax
open List
open Format

let parse_error s =
	printf "@[line %d:@ %s@]\n" !noline s ;
	Javasyntax.noline := 1
;;


type class_decl = 
	FIELDDECL of javafielddecl
  |	METHDEF of javamethdecl
;;

let rec process_cdecls cdecls =
  let rec pcdecls cdecls (fdecls, mdefs) =
  match cdecls
  with [] -> (fdecls,mdefs)
  |    FIELDDECL fd::cdecls' -> pcdecls cdecls' (fd::fdecls, mdefs)
  |    METHDEF   md::cdecls' -> pcdecls cdecls' (fdecls, md::mdefs)
  in pcdecls cdecls ([], [])
;;





type stmt =
	LOCDECL of javavardecl list
  |	STMT of javastat
;;

let rec process_bstmts bstmts =
  match bstmts
  with []		  -> [], []
  |    LOCDECL d::bstmts' -> let ldecls, stmts = process_bstmts bstmts'
			     in d@ldecls, stmts
  |    STMT s::bstmts'    -> let ldecls, stmts = process_bstmts bstmts'
			     in ldecls, s::stmts
;;

%}

%token <string> IDENT
%token TRUE FALSE
%token <string> CST_INT
%token <string> NAMED_TYPE

%token EOF
%token BOOL INT VOID NULL

%token EQ 
%token ARROW DOT

%token EQ_EQ EXCLAM_EQ INF SUP INF_EQ SUP_EQ
%token PLUS MINUS STAR SLASH PERCENT
%token EXCLAM AND_AND PIPE_PIPE

%token RPAREN LPAREN RBRACE LBRACE LBRACKET RBRACKET
%token COLON SEMICOLON COMMA 

%token CLASS EXTENDS


%token PUBLIC PROTECTED PRIVATE
%token STATIC FINAL 

%token IF ELSE
%token WHILE FOR
%token RETURN

%token NEW THIS SUPER

/* operator precedence */
%nonassoc 	IF
%nonassoc 	ELSE


%left	COMMA
%right	EQ 
%right	COLON
%left	PIPE_PIPE
%left	AND_AND
%left	EQ_EQ EXCLAM_EQ
%left	INF SUP INF_EQ SUP_EQ
%left	INF_INF SUP_SUP
%left	PLUS MINUS
%left	STAR SLASH 
%right	EXCLAM  RPAREN ADDROF
%left	DOT ARROW LPAREN LBRACKET LBRACE SIZEOF


/* Non-terminals informations */
%start file cdecl statement expression
%type <Javasyntax.javaprog> file
%type <Javasyntax.javaclassdecl> cdecl 
%type <Javasyntax.javastat> statement 
%type <Javasyntax.javaexp> expression


%%

/* ***********************************************************
 * Unite de compilation Java 
 * *********************************************************** */ 

file:
	tdecls   EOF	{JavaProg (rev $1)}
;

/* ***********************************************************
 * Type Declarations  
 * *********************************************************** */ 

tdecls:
	tdecl		{[$1]}
|	tdecls tdecl	{$2::$1}
|	tdecls tdecl SEMICOLON 	{$2::$1}
;

tdecl:
	cdecl		{ $1}
;


/* ***********************************************************
 * Declarations de classe
 * *********************************************************** */ 

cdecl:
	modifiers CLASS IDENT opt_super  LBRACE class_decls RBRACE 
			{let (fdecls, mdefs) = process_cdecls ($6)
		 	 in
		 	 {classname = $3; classfather = $4;
		    	  fielddecls = fdecls;
		    	  methdecls = mdefs}}
;

opt_super:
	/* empty */	{Nopt}
|	EXTENDS IDENT	{Some ($2)}
;



class_decls:
	/* empty */	{[]}
| 	class_decls class_decl	
			{$2@$1}
;

class_decl:
	class_member	{$1}
;

class_member:
	field_decl	{map (fun d -> FIELDDECL (d)) $1}
|	method_decl	{$1}
;

/* ***********************************************************
 * Declarations de champs (fields) et variables
 * *********************************************************** */ 

field_decl:
	modifiers jtype var_decls SEMICOLON
		{map (fun d ->  
                                {fieldname = fst d;
				fieldtype = $2;
				fieldprotection = fst $1;
                fieldstatus = snd $1;
				fieldinit = snd d})
                             $3}
;

var_decls:
	var_decl	{[$1]}
|       var_decls COMMA var_decl
			{$3::$1}
;

var_decl:
	IDENT	{$1, Nopt}
|	IDENT EQ expression	{$1, Some ($3)}
;


/* ***********************************************************
 * Declarations de methode
 * *********************************************************** */ 

method_decl:
	method_head method_body
			{let (protection,status) , restype, (name,params) = $1
			 in
			 [METHDEF ( 
			   	{methname = name; 
			   	 methrestype = restype;
			   	 methparams = params;
			   	 methprotection = protection;
			   	 methstatus = status;
			   	 methbody = $2 })]}
;
	
method_head:
	modifiers jtype method_dect
			{$1,$2,$3}
|	modifiers VOID method_dect
			{$1,JavaVoidType,$3}
;

method_dect:
	IDENT LPAREN method_params RPAREN
			{$1, $3}
|	IDENT LPAREN RPAREN
			{$1, []}
;

method_params:
	method_param	{[$1]}
|	method_params COMMA method_param	
			{$3::$1}
;

method_param:
	jtype IDENT
			{$2, $1}
;		


method_body:
	block		{$1}
|	SEMICOLON	{JavaBlock ([], [JavaReturn (Nopt)])}
;


/* ***********************************************************
 * Declarations de base
 * *********************************************************** */ 

modifiers:
	/* empty */	{Public,InstStatus}
|	modifiers PUBLIC
			{Public, snd $1}
|       modifiers PROTECTED
                        {Protected,snd $1}
|       modifiers PRIVATE
                        {Private,snd $1}
|       modifiers STATIC
                        {fst $1,ClassStatus}
;


jtype:
	prim_type	{$1}
|	ref_type	{$1}
;

prim_type:
	BOOL		{JavaBoolType}
|       INT     	{JavaIntType}
;

ref_type:
	class_inter_type	{$1}
;

class_inter_type:
	IDENT		{JavaIdType ($1)}
;


names:
	/* empty */	{[]}
|	IDENT   	{[$1]}
|	names COMMA IDENT
			{$3::$1}
;



local_decl:
	jtype var_decls SEMICOLON
		{map (fun d -> let vname,  vinit = d 
			       in
			       {varname = vname;
				vartype = $1;
				varinit = vinit})
                             $2}
;

/* ***********************************************************
 * Expressions
 * *********************************************************** */ 

expression:
	ass_expression	{$1}
;

ass_expression:
	binary_exp	{$1}
|	assignment	{$1}
;

binary_exp:
	unary_exp	{$1}
|	binary_exp PIPE_PIPE unary_exp
			{JavaBinopExp (JavaOR, $1, $3)}
|	binary_exp AND_AND unary_exp
			{JavaBinopExp (JavaAND, $1, $3)}
|	binary_exp EQ_EQ unary_exp
			{JavaBinopExp (JavaEGAL, $1, $3)}
|	binary_exp EXCLAM_EQ unary_exp
			{JavaBinopExp (JavaNE, $1, $3)}
|	binary_exp INF unary_exp
			{JavaBinopExp (JavaLT, $1, $3)}	
|	binary_exp SUP unary_exp
			{JavaBinopExp (JavaGT, $1, $3)}
|	binary_exp INF_EQ unary_exp
			{JavaBinopExp (JavaLE, $1, $3)}
|	binary_exp SUP_EQ unary_exp
			{JavaBinopExp (JavaGE, $1, $3)}
|	binary_exp PLUS unary_exp
			{JavaBinopExp (JavaPLUS, $1, $3)}
|	binary_exp MINUS unary_exp
			{JavaBinopExp (JavaMOINS, $1, $3)}
|	binary_exp STAR unary_exp
			{JavaBinopExp (JavaMULT, $1, $3)}
|	binary_exp SLASH unary_exp
			{JavaBinopExp (JavaDIV, $1, $3)}
;

unary_exp:
	IDENT		{JavaVarExp $1}	
|       literal         {JavaConstExp $1}
| 	objexp		{$1}
|       class_new       {$1}
|       field_acc       {$1}
|       method_inv      {$1}
|       array_acc       {$1}
|       array_new       {$1}
|	EXCLAM unary_exp
			{JavaUnopExp (JavaNOT, $2)}
;

assignment:
	lefthand EQ binary_exp
			{JavaAssign ($1, $3)}
;			

lefthand:
	IDENT		{JavaVarExp ($1)}
|	field_acc	{$1}
|	array_acc	{$1}
;

objexp:
	THIS		{JavaThisExp} 
|       IDENT           {JavaVarExp $1}
|	LPAREN expression RPAREN
			{$2}
;

class_new:
	NEW IDENT LPAREN RPAREN
			{JavaNewClassExp $2}
;

field_acc:
	objexp DOT IDENT
			{JavaFieldExp ($1, $3)}
;

method_inv:
	IDENT LPAREN opt_arguments RPAREN
			{JavaCall ($1, $3)}
|	objexp DOT IDENT LPAREN opt_arguments RPAREN
			{JavaMethCall ($1, $3, $5)}
;



literal:
	CST_INT		{JavaInt (int_of_string $1)}
|	TRUE		{JavaTrue}
|	FALSE		{JavaFalse}
|	NULL		{JavaNull}
;

opt_arguments:
	/* empty */	{[]}
|	arguments	{rev $1}
;

arguments:
	expression	{[$1]}
|	arguments COMMA expression	
			{$3::$1}
;

dim_exps:
	dim_exp		{[$1]}
|	dim_exps dim_exp
			{$2::$1}
;

dim_exp:
	LBRACKET expression RBRACKET	
			{$2}
;

dims:
	LBRACKET RBRACKET	
			{1}
|	dims LBRACKET RBRACKET
			{$1+1}
;

/* ***********************************************************
 * Blocks et instructions
 * *********************************************************** */ 

block:
	LBRACE bstmts RBRACE
                        {let ldecls, stmts = process_bstmts (rev $2)
			 in JavaBlock (ldecls, stmts)}
;
bstmts:
	bstmt		{[$1]}
|	bstmts bstmt 	{$2::$1}
;

bstmt:
	local_decl
			{LOCDECL ($1)}
|	statement	{STMT ($1)}
;

statement:
	block		{$1}
|	SEMICOLON	{JavaNop}
|	expression SEMICOLON
			{JavaComput $1}			
|	IF LPAREN expression RPAREN statement ELSE statement
			{JavaIf ($3, $5, $7)}
|	FOR LPAREN expression SEMICOLON expression 
            SEMICOLON expression RPAREN statement
			{JavaFor ($3, $5, $7, $9)}
|	RETURN SEMICOLON
			{JavaReturn (Nopt)}					
|	RETURN expression SEMICOLON
			{JavaReturn (Some ($2))}	
;


%%


