(*************************************************************
                        Mini Java
 *************************************************************
 Auteur :       MARTIN Ludovic
 Fichier :      typeur.ml
 *************************************************************
 Description :  typage de Mini Java en OCaml selon les regles 
                indiquees dans regles_typage.ps
 *************************************************************)


(*************************************************************
  Ouverture des modules utilises
 *************************************************************)
open Analyseur;;
open Javasyntax;;

(*************************************************************
  Declaration de l'exceptions levee en cas d'erreur de typage
 *************************************************************)
exception Erreur_typage of string;;


(*************************************************************
  Fonctions pour la definition de la relation de sous-typage
 *************************************************************)
(**
  Renvoie true si classname1 est une sousclasse de classname2.
   @param table resultat de @see table_ancetres ((string * string list) list)
   @param classname1 nom de la sous-classe (string)
   @param classname1 nom de la super-classe (string)
   @return     un boolean
 *)
let rec sousclasse table classname1 classname2 =
  List.mem classname2 (List.assoc classname1 table)
;;

(**
  Renvoie true si ty1 est un soustype de ty2. On considere que
  les inclusions proviennent de la hierarchie de classes et pas
  des types de base.
   @param table resultat de @see table_ancetres ((string * string list) list)
   @param ty1   nom du sous-type (javatype)
   @param ty2   nom du super-type (javatype)
   @return      un boolean
 *)
let rec soustype table  ty1 ty2 =
  (ty1=ty2) ||
  match (ty1,ty2)
  with     
    JavaIdType classname1 , JavaIdType "Object"
    -> true (* toute classe est sousclasse d'Object *)
  | JavaIdType classname1 , JavaIdType classname2
    -> sousclasse table classname1 classname2
  | JavaNullType , JavaIdType _
    -> true
  | _,_   -> false
;;

(*************************************************************
  Fonctions pour la construction des tables d'ancetres
  *************************************************************
  Pour les classes
 *************************************************************)
(**
  Calcule la definition de la classe Object
  *)
let object_class = cdecl_of_string "class Object{}";;

(* Declaration de mthode bidon pour getclassdef :  effacer *)
let classdecl_bidon = 
  { classname = "bidon"; 
    classfather = Nopt; 
    fielddecls = [];
    methdecls = [] }
;;

(**
  Renvoie la definition d'une classe. Leve une exception si la classe 
  n'existe pas.
   @param cname nom de la classe (string)
   @param prog  programme a typer (javaprog)
   @return      (methdecl)
  *)
let getclassdef cname (JavaProg(classlist)) = 
  if cname = "Object" then object_class
  else let pred = fun d -> d.classname = cname in 
       List.find pred classlist
;;

(**
  Calcule la liste des ancetres pour la classe dont le nom est 
  donn en parametre. Leve exception si heritage recursif.
   @param cname nom de la classe (string)
   @param prog  programme a typer (javaprog)
   @return      liste des ancetres (string list)
 *)
exception Classes_mutuellement_rec ;;

let  ancetres cname prog = 
  let rec anc c p dv = 
    let classe = getclassdef c p in
    match classe.classfather
      with Nopt -> ["Object"]
      | Some(pere) -> if List.mem pere dv then raise Classes_mutuellement_rec
                      else pere::(anc pere p (classe.classname::dv))
  in anc cname prog []
;;

(**
  Calcule la table d'ancetres pour toutes les classes du 
  programme.
   @param prog  programme a typer (javaprog)
   @return      table des ancetres, pour chaque classe
                sa liste d'ancetres ((string * string list) list)
 *)
let table_ancetres (JavaProg(classlist) as prog) =
  List.map 
    (function {classname=c} -> c, ancetres c prog)
    classlist
;;

let father cname table = List.hd (List.assoc cname table);;


(*************************************************************
  Fonctions pour la rsolution de la surcharge a l'appel de methode
 *************************************************************)
(**
  Renvoie true si la liste des arguments a l'appel est
  compatible (par sous-typage) avec la liste des parametres, 
  c'est-a-dire :
  - la liste des arguments a l'appel a la mm longueur que celle des 
    parametres formels
  - chaque type argument est le sous-type de son correspondant dans la
    liste des parametres
   @param table      resultat de @see table_ancetres ((string * string list) list)
   @param argtypes   liste des types des arguments (javatype list)
   @param paramtypes liste des types des parametres formels  (javatype list)
   @return           boolean
 *)
let signature_acceptable table argtypes paramtypes =
  try
    List.for_all (function (x,y) -> soustype table  x y)
                 (List.combine argtypes paramtypes)
  with 
    Invalid_argument "List.combine"
    -> false (* car l'exception est levee si les deux listes n'ont pas la mm longueur *)
;;

(**
  Renvoie la liste des types des parametres formels d'une methode a
  partir de la definition de la methode.
   @param methdef la definition de la methode (javamethdecl)
   @return        liste de types des parametres formels (javatype list)
 *)
let signature methdef =  
  List.map snd methdef.methparams
;;

(**
  Cacule la liste des methodes qui sont definies dans la hierarchie d'une classe
  et qui ont le mm nom et les parametres compatibles avec les arguments.
   @param prog       programme type (javaprog)
   @param table      resultat de @see table_ancetres ((string * string list) list)
   @param mname      nom de la methode cherchee (string)
   @param argtypes   liste des types des arguments d'appel (javatype list)
   @param cname      nom de la classe d'appel (string)
   @return           liste (classe de definition, methode)
                     (string * javamethdecl) list
 *)
let methodes_compatibles prog table mname argtypes cname =
  (* on construit la hierarchie de classes dans lesquelles on cherche la methode *)
  let hierarchie = cname::List.assoc cname table in
  (* on definit une fonction locale renvoie les methodes compatibles
     dans la classe c *)
  let methodes_compatibles_dans_classe c =
    let methlist =
          List.filter 
             (function {methname=mn} as methdef
               -> mn=mname &&  signature_acceptable table argtypes (signature methdef))
             ((getclassdef c prog).methdecls)
     in List.map (function meth -> c, meth) methlist
   in
    List.flatten 
    (List.map methodes_compatibles_dans_classe  hierarchie)
;;


(**
  Calcule la definition d'une methode, la plus specifique par rapport
  a la classe demandee a l'appel et aux arguments. Leve une exception
  si la definition n'as pas ete trouvee.
   @param prog       programme type (javaprog)
   @param table      resultat de @see table_ancetres ((string * string list) list)
   @param classname  nom de la classe d'appel (string)
   @param methname   nom de la methode appelee (string)
   @param argtypes   liste des types des arguments d'appel (javatype list)
   @return           la definition de la methode (javamethdecl)
 *)
let trouver_methode prog table classname methname argtypes =
  (* on construit la liste des methodes compatibles dans la hierarchie de classname *)
  let methodes = methodes_compatibles prog table methname argtypes classname in
  try
    (* on cherche dans methodes celle qui est la plus spcifique, cad,
       celle dont la liste classe de dfinition::signature est un sous-type
       (@see signature_acceptable) par toute les autres methodes.
     *)
    snd (List.find
      (function (clname,methdef) ->
        List.for_all
          (function (clname1,methdef1) -> 
	          signature_acceptable table
              (JavaIdType clname::signature methdef)
              (JavaIdType clname1::signature methdef1))
          methodes
      )
     methodes)
  with
    Not_found -> raise(Erreur_typage("La methode " ^ methname ^ 
                           " n'existe pas ou est definie de facon ambigue" ^
                           "dans la classe " ^ classname))
;;

(*************************************************************
  Fonctions utiles pour detecter les doublons
 *************************************************************)
 (* Verifie qu'une liste ne contient pas de doublon *)
 (* Sert a verifier l'unicite des champs dans une classe *)
 (* et des parametres dans une methode *)
 let is_set l =
    List.for_all 
	  (function id -> 
	    (List.length(List.filter (function id1 -> id=id1) l) = 1)
	  )
	  l;;
	  
(* Liste des methodes definies dans la hierarchie de la classe cname *)
(* ayant un nom et une signature donnees *)
(* Retourne javamethdecl list *)
let methodes_de_meme_signature prog table mname signature cname =
  (* on construit l'hierarchie de classes dans laquel on cherche la methode *)
  let hierarchie = cname::List.assoc cname table in
  (* on definit une fonction locale renvoie les methodes de signature "signature"
     dans la classe cname *)
  let methodes_de_meme_signature_dans_classe mname argtypes cname =
    List.filter (function md -> md.methname = mname
                             && List.map snd md.methparams = signature) 
                (getclassdef cname prog).methdecls in
    List.flatten 
    (List.map (methodes_de_meme_signature_dans_classe mname signature) hierarchie)
;;

(* Sert a verifier que lorsqu'on redefinit une methode avec la meme signature *)
(* le type resultat n'est pas change *)
(* Retourne boolean *)
let resultat_unique prog table mname signature cname =
 let methlist = methodes_de_meme_signature prog table mname signature cname in
 let reslist = List.map (function md -> md.methrestype) methlist in
   match reslist
   with ([]|[_]) -> true
     |  a::l -> List.for_all (function x -> x=a) l;;
     
	  
(*************************************************************
  Typage d'une expression
 *************************************************************)
 let typer_unopexp op ty =
     match (op,ty)
     with (JavaUMOINS,JavaIntType) ->  JavaIntType
      | (JavaNOT,JavaBoolType)     ->  JavaBoolType
      |  _ -> raise (Erreur_typage("Mauvais argument pour operateur unaire"))  
      ;;
 
 
  let typer_binopexp op ty1 ty2 =
     match (op,ty1,ty2)
     with JavaAND,JavaBoolType,JavaBoolType -> JavaBoolType
      | JavaOR,JavaBoolType,JavaBoolType    -> JavaBoolType
      | JavaPLUS,JavaIntType,JavaIntType    -> JavaIntType
      | JavaMOINS,JavaIntType,JavaIntType   -> JavaIntType
      | JavaMULT,JavaIntType,JavaIntType    -> JavaIntType
      | JavaDIV,JavaIntType,JavaIntType     -> JavaIntType
      | JavaEGAL,JavaIntType,JavaIntType    -> JavaBoolType
      | JavaEGAL,JavaBoolType,JavaBoolType  -> JavaBoolType
      | JavaNE,JavaIntType,JavaIntType      -> JavaBoolType
      | JavaNE,JavaBoolType,JavaBoolType    -> JavaBoolType
      | JavaLT,JavaIntType,JavaIntType      -> JavaBoolType
      | JavaGT,JavaIntType,JavaIntType      -> JavaBoolType
      | JavaLE,JavaIntType,JavaIntType      -> JavaBoolType
      | JavaGE,JavaIntType,JavaIntType      -> JavaBoolType
      |  _ -> raise (Erreur_typage("Mauvais argument pour operateur binaire")) 
      ;;

(* Est-ce que l'expression  gauche peut tre affecte ? *)     

let is_assignable e =
  match e
  with (JavaVarExp _ | JavaFieldExp _) -> true
   |     _  -> false;;

(* Renvoie la dfinition du champ et le nom de la classe dans laquelle 
   apparait le nom du champ *)
   
let get_field_def prog table cname fname =
   let hierarchie = cname::List.assoc cname table in
     let rec get fname l =
     match l with []  ->  raise Not_found
       |  c::l' -> (try let fd = List.find (function fd -> fd.fieldname=fname) 
                                            (getclassdef c prog).fielddecls
                        in fd,c
                    with Not_found -> get fname l')
     in get fname hierarchie;;
     

     
(*
  Typage d'une expression : cf les regles de typage.
   @param prog   programme type (javaprog)
   @param table  table d'heritage ((string * string list) list)
   @param cname  classe de definition (string)
   @param ctx    contexte de typage pour les variables ((string * javatype) list)
   @param status statut de la methode dans laquelle apparait l'expr
   @param exp    expression (javaexp)
   @return       type de l'expression (javatype)
 *)
let rec typer_exp prog table cname ctx status exp = 
  match exp with
    JavaConstExp (JavaInt _)            -> JavaIntType
  | JavaConstExp (JavaTrue | JavaFalse) -> JavaBoolType
  | JavaConstExp JavaNull -> JavaNullType
  | JavaThisExp -> (
      if status = InstStatus then
        JavaIdType cname
      else
        raise (Erreur_typage("Evaluation d'une var d'instance dans une meth de classe")))
  | JavaVarExp v -> (
      (* cas d'une var locale *)
      try 
        List.assoc v ctx
      with
        _ ->
          (try
            let champ = fst (get_field_def prog table cname v) in
            (* cas d'un champ d'inst *)
            if status = InstStatus then
              if champ.fieldstatus = InstStatus then champ.fieldtype
              else raise (Erreur_typage("champ de classe dans une meth d'instance"))
            else
              if champ.fieldstatus = ClassStatus then champ.fieldtype
              else raise (Erreur_typage("champ d'instance dans une meth de classe"))
          with
            Not_found -> raise (Erreur_typage("Var "^v^" non declaree"))))
  | JavaNewClassExp id -> (
      try
        let classe = getclassdef id prog in
        JavaIdType classe.classname
      with
        Not_found -> raise (Erreur_typage("Classe "^id^" non declaree new")))
  | JavaFieldExp(e, fname) -> (
      let liste_champ = fun nom_classe -> 
        (let hierarchie = nom_classe::List.assoc nom_classe table in
        let classtofield = fun c_nom -> (getclassdef c_nom prog).fielddecls in
        List.flatten (List.map classtofield hierarchie)) in
      match typer_exp prog table cname ctx status e with
        JavaIdType id -> 
          (try
            let classe = getclassdef id prog in
            (try 
              let champdecl = List.find (fun f->f.fieldname=fname) (liste_champ classe.classname) in
              champdecl.fieldtype
            with
              Not_found ->raise (Erreur_typage(fname^" n'est pas un champ de "^
                                               classe.classname)))
          with
            Not_found ->
              raise (Erreur_typage("Impossible de retrouve a quelle classe est associe le champ "
                                    ^fname)))
      | _ -> raise (Erreur_typage(fname^" n'est pqs associe a une classe")))
  | JavaMethCall(e, mname, explist) -> (
      match typer_exp prog table cname ctx status e with
        JavaIdType id -> 
          (try
            (* Classe a la quelle appartient la methode appelee *)
            let classe = getclassdef id prog in 
            (* Liste des types des arguments de la methode *)
            let typelist = List.map (fun e -> typer_exp prog table cname ctx status e) explist in
            (* declaration de la methode appelee *)
            let methode_decl = trouver_methode prog table classe.classname mname typelist in
	          methode_decl.methrestype
          with
            Not_found -> 
              raise (Erreur_typage("Impossible de retrouve a quelle classe est associe la methode "
                                    ^mname)))
      |	_ -> raise (Erreur_typage(mname^" n'est pas associe a une classe")))
  | JavaCall(mname, explist) -> (
      (* Liste des types des arguments de la methode *)
      let typelist = List.map (fun e -> typer_exp prog table cname ctx status e) explist in
      (* declaration de la methode appelee *)
      let methode_decl = trouver_methode prog table cname mname typelist in
      methode_decl.methrestype)
  | JavaUnopExp(op, e)       ->   
      typer_unopexp op (typer_exp prog table cname ctx status e)
  | JavaBinopExp(op, e1, e2) ->               
      typer_binopexp   op  (typer_exp prog table cname ctx status e1)
                          (typer_exp prog table cname ctx status e2)
  | JavaAssign(e1, e2) -> (
      let type1 = typer_exp prog table cname ctx status e1 in
      let type2 = typer_exp prog table cname ctx status e2 in
      if (soustype table type2 type1) then
	      if (is_assignable e1) then
	        type1
	      else
	        raise (Erreur_typage("L'exp a gauche de = n'est pas assignable"))
      else
	      raise (Erreur_typage("L'exp a droite de = n'est pas sous type de celle de gauche")))
;;

(*************************************************************
  Typage d'une declaration de variable
 *************************************************************)
(**
  Typage d'une declaration de variables :
   - le type de declaration et celui de l'expression (quand il y a)
   doivent correspondre.
   @param prog   programme type (javaprog)
   @param table  table d'heritage ((string * string list) list)
   @param name   classe de definition (string)
   @param ctx    contexte de typage pour les variables ((string * javatype) list)
   @param status statut de la methode dans laquelle apparait l'expr
   @param vardecl declaration a typer (javavardecl)
   @return       le nouvel contexte ((string * javatype) list)
 *)
let typer_vardecl prog table cname ctx status vardecl =
  (* on verifie que la variable n'est pas deja dans le contexte *)
  if List.mem vardecl.varname (List.map fst ctx) then
    raise (Erreur_typage("La variable "^vardecl.varname^" existe deja"))
  else
    match vardecl.varinit with
      Nopt -> 
	      (vardecl.varname,vardecl.vartype)::ctx
    | Some e -> 
        (* on calcule le type de l'expression d'initialisation *)
	      if soustype table (typer_exp prog table cname ctx status e) vardecl.vartype then 
          (* on verifie que ce type est un sous-type du type declare *)
          (vardecl.varname,vardecl.vartype)::ctx
	      else
          raise (Erreur_typage("L'initialisation de la var "^vardecl.varname^
                                " n'est pas un sous type de sa declaration"))
;;

(* on applique recoursivement la fonction ci-dessus *)
let rec typer_vardecllist prog table cname ctx status varlist =
  match varlist with 
    [] -> ctx
  | vardecl::l ->
      let newctx = typer_vardecl prog table cname ctx status vardecl in
      typer_vardecllist prog table cname newctx status l
;;
 

(*************************************************************
  Typage d'une instruction
 *************************************************************)

(**
  Typage d'une instruction : cf regles de typage
   @param prog   programme type (javaprog)
   @param table  table d'heritage ((string * string list) list)
   @param cname  classe de definition (string)
   @param ctx    contexte de typage pour les variables ((string * javatype) list)
   @param status statut de la methode dans laquelle apparait l'instruction
   @param stat   instruction a typer (javastat)
   @return       rien (unit)
 *)
let rec typer_stat prog table cname ctx status stat =
  match stat with
    JavaBlock (varlist, statlist) ->
      (* Augmente le context avec les declarations de var *)
      let rec add_ctx = fun liste_var env ->
        match liste_var with
          [] -> env
        | var::l -> add_ctx l (typer_vardecl prog table cname env status var) in
      let new_ctx = add_ctx varlist ctx in
      (* Verifie les instructions du bloc *)
      if (List.for_all (fun x -> x=())
         (List.map (fun s->typer_stat prog table cname new_ctx status s) statlist))
      then ()
  | JavaComput e ->
      (* Verifie que e est une expression valide *)
      let _ = typer_exp prog table cname ctx status e
      in ()
  | JavaIf(e, sthen, selse) ->
      (* Verifie que la condition est du type boolean *)
      if typer_exp prog table cname ctx status e = JavaBoolType then
        (* Verifie les instructions du then et du else *)
        if typer_stat prog table cname ctx status sthen = () then
          if typer_stat prog table cname ctx status selse = () then
            ()
  | JavaFor(einit, econd, estep, s) ->
      (* Verifie que l'initialisation est correctement typee *)
      let _ = typer_exp prog table cname ctx status einit in
      (* Verifie que la condition est du type boolean *)
      if typer_exp prog table cname ctx status econd = JavaBoolType then
        (* Verifie que l'instruction d'iteration est correctement typee *)
        let _ = typer_exp prog table cname ctx status estep in
        (* Verifie le bloc du for *)
        if typer_stat prog table cname ctx status s = () then
          ()
  | JavaReturn(Some e) ->
      let type_ret = List.assoc "return" ctx in
      let type_exp = typer_exp prog table cname ctx InstStatus e in
      if soustype table type_exp type_ret
      then ()
      else raise (Erreur_typage("")) (* Erreur interceptee plus loin *)
  | JavaReturn(Nopt) ->
      if List.assoc "return" ctx = JavaVoidType 
      then ()
      else raise (Erreur_typage("")) (* Erreur interceptee plus loin *) 
  | JavaNop -> ()
;;

(*************************************************************
  Typage d'une declaration de champ
 *************************************************************)
(**
  Typage d'une declaration de champ :
   - le type de declaration et celui de l'expression (quand il y a)
   doivent correspondre.
   @param prog   programme type (javaprog)
   @param table  table d'heritage ((string * string list) list)
   @param cname   classe de definition (string)
   @param fielddecl declaration a typer (javafielddecl)
   @return       rien (unit)
 *)
let typer_fielddecl prog table cname fielddecl =
  (* Verifie qu'un champ de meme nom n'existe pas deja ds la classe *)
  if (is_set (List.map (fun c -> c.fieldname) (getclassdef cname prog).fielddecls)) then
    (* Regarde si le champ est initialise *)
    match fielddecl.fieldinit with
      Nopt -> ()
    | Some e -> 
        (* Typage de l'expression d'initialisaion et verif de correspondance des types *)
        if soustype table (typer_exp prog table cname [] fielddecl.fieldstatus e) fielddecl.fieldtype
        then 
          ()
        else
          raise (Erreur_typage("L'initialisation du champ "^fielddecl.fieldname^
                                " n'est pas un sous type de sa declaration"))
  else
    raise (Erreur_typage("Le champ "^fielddecl.fieldname^" est deja declare ds "^cname))
;;


(*************************************************************
  Typage d'une definition de methode
 *************************************************************)
(**
  Typage d'une declaration de methode : 
   - les noms des parametres doivent etre differents
   - le type du resultat et celui de l'instruction (quand il y a)
   doivent correspondre.
   @param prog   programme type (javaprog)
   @param table  table d'heritage ((string * string list) list)
   @param cname  classe  de definition (string)
   @param methdecl declaration a typer (javamethdecl)
   @return       rien (unit)
 *)
let typer_methdecl prog table cname methdecl =
  (* on verifie que les noms des parametres sont differents *)
  if is_set (List.map fst methdecl.methparams) then
    (* Verifie que le type retourne et celui du corps sont les mm *)
    let ctx = ("return",methdecl.methrestype)::methdecl.methparams in
    let res = 
      (try
        typer_stat prog table cname ctx methdecl.methstatus methdecl.methbody
      with
        Erreur_typage "" -> raise (Erreur_typage("Le type de retour de "^methdecl.methname^
                                             " et celui de son corps ne sont pas les mm"))) in
    (* Verifie que 2  methodes de meme nom et meme signature n'existent pas ds la meme classe *)
    let mini_table = (cname,[])::[] in
    let methlist = 
      methodes_de_meme_signature prog mini_table methdecl.methname (signature methdecl) cname in
    if List.length methlist > 1 then
      raise (Erreur_typage("La meth "^methdecl.methname^
                           " est definie plusieurs fois ds "^cname))
    else
      (* Verifie si il existe ds les classes sup des methodes de meme nom et meme signature *)
      (* Ces methodes doivent toutes retourner le meme type *)
      if resultat_unique prog table methdecl.methname (signature methdecl) cname then
        (* Verifie le status de la methode declaree *)
        (* Une methode ne peut etre redefinie que par une methode de meme status *)
        let methlist2 = 
          methodes_de_meme_signature prog table methdecl.methname (signature methdecl) cname in
        if List.for_all (fun md -> md.methstatus = InstStatus) methlist2 then
          (* Toutes les methodes definies sont d'instance *)
          ()
        else
          if List.for_all (fun md -> md.methstatus = ClassStatus) methlist2 then
            (* Toutes les methodes definies sont de classe *)
            ()
          else
            if methdecl.methstatus = ClassStatus then
              raise (Erreur_typage(cname^"."^methdecl.methname^
                                    " redefinit une methode d'instance"))
            else
              raise (Erreur_typage(cname^"."^methdecl.methname^
                                    " redefinit une methode de classe"))
      else
        raise (Erreur_typage("La methode "^methdecl.methname^
                              " recouvre une methode de type different"))
  else
    raise (Erreur_typage("La methode "^methdecl.methname^" a plusieurs param de meme nom"))
;;



(*************************************************************
  Typage des declarations de classe
 *************************************************************)

(**
  Typage d'une declaration de classe : 
   @param prog   programme type (javaprog)
   @param table  table d'heritage ((string * string list) list)
   @param cdecl  declaration a typer (javainterdecl)
   @return       bool
 *)
let typer_classdef (JavaProg(classlist) as prog) table cdecl =
  (* le champ classfather a ete deja verifie par @see table_ancetres *)
  (* on verifie les definitions de champs *)
  (* On verifie d'abord qu'un champ n'est pas defini deux fois *)
  let fnames = List.map (function fd -> fd.fieldname) cdecl.fielddecls in
  if not (is_set fnames) then
    raise(Erreur_typage("Un champ est defini plusieur fois dans la classe " ^ cdecl.classname))
  else
    let fields = 
    (List.map 
       (function fielddecl -> typer_fielddecl prog table cdecl.classname fielddecl)
       cdecl.fielddecls
    ) in 
    (* on verifie les definitions de methodes *)
    let methods = 
      (List.map
         (function methdecl -> typer_methdecl prog table cdecl.classname methdecl)
         cdecl.methdecls
      ) in
    true
;;


(*************************************************************
  Typage d'un programme
 *************************************************************)
(**
  Typage d'un programme par typage des declarations d'interfaces 
  et de classes. Les fonctions qu'elle appelle peuvent lever
  une exception Erreur_typage.
   @param prog       programme type (javaprog)
   @return           bool
 *)
let typer_prog (JavaProg(classlist) as prog) =
  (* on construit la table de la hierarchie des classes *)
  (* cette construction verifie en mm temps que l'heritage est correct,
     cad, que chaque nom de classe  est defini et
     que l'heritage n'est pas recursif *)
  let  table_c = table_ancetres prog in
   List.for_all
     (function classdef -> typer_classdef prog table_c classdef)
     classlist
;;
